div.sphx-glr-download-link-note {
    height: 0px;
    visibility: hidden;
}

p.sphx-glr-timing {
    height: 0px;
    visibility: hidden;
}

div.sphx-glr-download {
    height: 0px;
    visibility: hidden;
}

p.sphx-glr-signature {
    height: 0px;
    visibility: hidden;
}

dl.field-list .colon {
    display: none;
}
